@@ -13,13 +13,17 @@
 #include "keycache.h"
 #include "keycache_p.h"
 
-#include "libkleo_debug.h"
-
+#include "kleo/keygroup.h"
 #include "kleo/predicates.h"
 #include "kleo/stl_util.h"
 #include "kleo/dn.h"
+
+#include "utils/compat.h"
 #include "utils/filesystemwatcher.h"
 
+#include <KConfigGroup>
+#include <KSharedConfig>
+
 #include <gpgme++/error.h>
 #include <gpgme++/context.h>
 #include <gpgme++/key.h>
@@ -45,12 +49,17 @@
 #include <functional>
 #include <iterator>
 
+#include "kleo/debug.h"
+#include "libkleo_debug.h"
+
 using namespace Kleo;
 using namespace GpgME;
 using namespace KMime::Types;
 
 static const unsigned int hours2ms = 1000 * 60 * 60;
 
+static const QString groupNamePrefix = QStringLiteral("Group-");
+
 //
 //
 // KeyCache
@@ -62,6 +71,18 @@ namespace
 
 make_comparator_str(ByEMail, .first.c_str());
 
+QStringList getFingerprints(const KeyGroup::Keys &keys)
+{
+    QStringList fingerprints;
+    fingerprints.reserve(keys.size());
+    std::transform(keys.cbegin(), keys.cend(),
+                    std::back_inserter(fingerprints),
+                    [] (const Key &key) {
+                        return QString::fromLatin1(key.primaryFingerprint());
+                    });
+    return fingerprints;
+}
+
 }
 
 class KeyCache::Private
@@ -182,11 +203,23 @@ public:
 
     void ensureCachePopulated() const;
 
-    void updateGroupCache() {
-        // Update Group Keys
-        // this is a quick thing as it only involves reading the config
-        // so no need for a job.
+    std::vector<Key> getKeysForGroup(const QStringList &fingerprints) const
+    {
+        std::vector<Key> groupKeys;
+        groupKeys.reserve(fingerprints.size());
+        for (const QString &fpr : fingerprints) {
+            const Key key = q->findByFingerprint(fpr.toLatin1().constData());
+            if (key.isNull()) {
+                qCDebug (LIBKLEO_LOG) << "Ignoring unknown key with fingerprint:" << fpr;
+                continue;
+            }
+            groupKeys.push_back(key);
+        }
+        return groupKeys;
+    }
 
+    void readGroupsFromGpgConf()
+    {
         // According to Werner Koch groups are more of a hack to solve
         // a valid usecase (e.g. several keys defined for an internal mailing list)
         // that won't make it in the proper keylist interface. And using gpgconf
@@ -196,30 +229,231 @@ public:
             return;
         }
 
-        auto entry = conf->entry(QStringLiteral("gpg"),
-                                 QStringLiteral("Configuration"),
-                                 QStringLiteral("group"));
+        auto entry = getCryptoConfigEntry(conf, "gpg", "group");
         if (!entry) {
             return;
         }
 
-        m_groups.clear();
-        for (const auto &value: entry->stringValueList()) {
-            const auto split = value.split(QLatin1Char('='));
+        // collect the key fingerprints for all groups read from the configuration
+        QMap<QString, QStringList> fingerprints;
+        for (const QString &value: entry->stringValueList()) {
+            const QStringList split = value.split(QLatin1Char('='));
             if (split.size() != 2) {
                 qCDebug (LIBKLEO_LOG) << "Ignoring invalid group config:" << value;
                 continue;
             }
-            const auto name = split[0];
-            const auto key = q->findByFingerprint(split[1].toLatin1().constData());
-            if (key.isNull()) {
-                qCDebug (LIBKLEO_LOG) << "Ignoring unknown fingerprint:" << split[1] << "in group" << name;
-                continue;
-            }
-            auto vec = m_groups.value(name);
-            vec.push_back(key);
-            m_groups.insert(name, vec);
+            const QString groupName = split[0];
+            const QString fingerprint = split[1];
+            fingerprints[groupName].push_back(fingerprint);
         }
+
+        // add all groups read from the configuration to the list of groups
+        for (auto it = fingerprints.cbegin(); it != fingerprints.cend(); ++it) {
+            const QString groupName = it.key();
+            const std::vector<Key> groupKeys = getKeysForGroup(it.value());
+            KeyGroup g(groupName, groupName, groupKeys, KeyGroup::GnuPGConfig);
+            m_groups.push_back(g);
+        }
+    }
+
+    KeyGroup readGroupFromGroupsConfig(const QString &groupId) const
+    {
+        Q_ASSERT(!m_groupsConfigName.isEmpty());
+
+        KSharedConfigPtr groupsConfig = KSharedConfig::openConfig(m_groupsConfigName);
+        const KConfigGroup configGroup = groupsConfig->group(groupNamePrefix + groupId);
+
+        const QString groupName = configGroup.readEntry("Name", QString());
+        const QStringList fingerprints = configGroup.readEntry("Keys", QStringList());
+        const std::vector<Key> groupKeys = getKeysForGroup(fingerprints);
+
+        // treat group as immutable if any of its entries is immutable
+        const QStringList entries = configGroup.keyList();
+        const bool isImmutable = configGroup.isImmutable() || std::any_of(entries.begin(), entries.end(),
+                                                                          [configGroup] (const QString &entry) {
+                                                                              return configGroup.isEntryImmutable(entry);
+                                                                          });
+
+        KeyGroup g(groupId, groupName, groupKeys, KeyGroup::ApplicationConfig);
+        g.setIsImmutable(isImmutable);
+        qCDebug(LIBKLEO_LOG) << "Read group" << g;
+
+        return g;
+    }
+
+    void readGroupsFromGroupsConfig()
+    {
+        if (m_groupsConfigName.isEmpty()) {
+            qCDebug(LIBKLEO_LOG) << "readGroupsFromGroupsConfig - name of application configuration file is unset";
+            return;
+        }
+
+        const KSharedConfigPtr groupsConfig = KSharedConfig::openConfig(m_groupsConfigName);
+        const QStringList configGroups = groupsConfig->groupList();
+        for (const QString &configGroupName : configGroups) {
+            qCDebug(LIBKLEO_LOG) << "Reading config group" << configGroupName;
+            if (configGroupName.startsWith(groupNamePrefix)) {
+                const QString keyGroupId = configGroupName.mid(groupNamePrefix.size());
+                if (keyGroupId.isEmpty()) {
+                    qCWarning(LIBKLEO_LOG) << "Config group" << configGroupName << "has empty group id";
+                    continue;
+                }
+                KeyGroup group = readGroupFromGroupsConfig(keyGroupId);
+                m_groups.push_back(group);
+            }
+        }
+    }
+
+    KeyGroup writeGroupToGroupsConfig(const KeyGroup &group)
+    {
+        if (m_groupsConfigName.isEmpty()) {
+            qCDebug(LIBKLEO_LOG) << "writeGroupToGroupsConfig - name of application configuration file is unset";
+            return KeyGroup();
+        }
+
+        Q_ASSERT(!group.isNull());
+        Q_ASSERT(group.source() == KeyGroup::ApplicationConfig);
+        if (group.isNull() || group.source() != KeyGroup::ApplicationConfig) {
+            qCDebug(LIBKLEO_LOG) << "writeGroupToGroupsConfig - group cannot be written to application configuration:" << group;
+            return KeyGroup();
+        }
+
+        KSharedConfigPtr groupsConfig = KSharedConfig::openConfig(m_groupsConfigName);
+        KConfigGroup configGroup = groupsConfig->group(groupNamePrefix + group.id());
+
+        qCDebug(LIBKLEO_LOG) << "Writing config group" << configGroup.name();
+        configGroup.writeEntry("Name", group.name());
+        configGroup.writeEntry("Keys", getFingerprints(group.keys()));
+
+        // reread group to ensure that it reflects the saved group in case of immutable entries
+        return readGroupFromGroupsConfig(group.id());
+    }
+
+    bool removeGroupFromGroupsConfig(const KeyGroup &group)
+    {
+        if (m_groupsConfigName.isEmpty()) {
+            qCDebug(LIBKLEO_LOG) << "removeGroupFromGroupsConfig - name of application configuration file is unset";
+            return false;
+        }
+
+        Q_ASSERT(!group.isNull());
+        Q_ASSERT(group.source() == KeyGroup::ApplicationConfig);
+        if (group.isNull() || group.source() != KeyGroup::ApplicationConfig) {
+            qCDebug(LIBKLEO_LOG) << "removeGroupFromGroupsConfig - group cannot be removed from application configuration:" << group;
+            return false;
+        }
+
+        KSharedConfigPtr groupsConfig = KSharedConfig::openConfig(m_groupsConfigName);
+        KConfigGroup configGroup = groupsConfig->group(groupNamePrefix + group.id());
+
+        qCDebug(LIBKLEO_LOG) << "Removing config group" << configGroup.name();
+        configGroup.deleteGroup();
+
+        return true;
+    }
+
+    void updateGroupCache()
+    {
+        // Update Group Keys
+        // this is a quick thing as it only involves reading the config
+        // so no need for a job.
+
+        m_groups.clear();
+        if (m_groupsEnabled) {
+            readGroupsFromGpgConf();
+            readGroupsFromGroupsConfig();
+        }
+    }
+
+    bool insert(const KeyGroup &group)
+    {
+        Q_ASSERT(!group.isNull());
+        Q_ASSERT(group.source() == KeyGroup::ApplicationConfig);
+        if (group.isNull() || group.source() != KeyGroup::ApplicationConfig) {
+            qCDebug(LIBKLEO_LOG) << "KeyCache::Private::insert - Invalid group:" << group;
+            return false;
+        }
+        const auto it = std::find_if(m_groups.cbegin(), m_groups.cend(),
+                                    [group] (const auto &g) {
+                                        return g.source() == group.source() && g.id() == group.id();
+                                    });
+        if (it != m_groups.cend()) {
+            qCDebug(LIBKLEO_LOG) << "KeyCache::Private::insert - Group already present in list of groups:" << group;
+            return false;
+        }
+
+        const KeyGroup savedGroup = writeGroupToGroupsConfig(group);
+        if (savedGroup.isNull()) {
+            qCDebug(LIBKLEO_LOG) << "KeyCache::Private::insert - Writing group" << group.id() << "to config file failed";
+            return false;
+        }
+
+        m_groups.push_back(savedGroup);
+
+        Q_EMIT q->groupAdded(savedGroup);
+
+        return true;
+    }
+
+    bool update(const KeyGroup &group)
+    {
+        Q_ASSERT(!group.isNull());
+        Q_ASSERT(group.source() == KeyGroup::ApplicationConfig);
+        if (group.isNull() || group.source() != KeyGroup::ApplicationConfig) {
+            qCDebug(LIBKLEO_LOG) << "KeyCache::Private::update - Invalid group:" << group;
+            return false;
+        }
+        const auto it = std::find_if(m_groups.cbegin(), m_groups.cend(),
+                                    [group] (const auto &g) {
+                                        return g.source() == group.source() && g.id() == group.id();
+                                    });
+        if (it == m_groups.cend()) {
+            qCDebug(LIBKLEO_LOG) << "KeyCache::Private::update - Group not found in list of groups:" << group;
+            return false;
+        }
+        const auto groupIndex = std::distance(m_groups.cbegin(), it);
+
+        const KeyGroup savedGroup = writeGroupToGroupsConfig(group);
+        if (savedGroup.isNull()) {
+            qCDebug(LIBKLEO_LOG) << "KeyCache::Private::update - Writing group" << group.id() << "to config file failed";
+            return false;
+        }
+
+        m_groups[groupIndex] = savedGroup;
+
+        Q_EMIT q->groupUpdated(savedGroup);
+
+        return true;
+    }
+
+    bool remove(const KeyGroup &group)
+    {
+        Q_ASSERT(!group.isNull());
+        Q_ASSERT(group.source() == KeyGroup::ApplicationConfig);
+        if (group.isNull() || group.source() != KeyGroup::ApplicationConfig) {
+            qCDebug(LIBKLEO_LOG) << "KeyCache::Private::remove - Invalid group:" << group;
+            return false;
+        }
+        const auto it = std::find_if(m_groups.cbegin(), m_groups.cend(),
+                                    [group] (const auto &g) {
+                                        return g.source() == group.source() && g.id() == group.id();
+                                    });
+        if (it == m_groups.cend()) {
+            qCDebug(LIBKLEO_LOG) << "KeyCache::Private::remove - Group not found in list of groups:" << group;
+            return false;
+        }
+
+        const bool success = removeGroupFromGroupsConfig(group);
+        if (!success) {
+            qCDebug(LIBKLEO_LOG) << "KeyCache::Private::remove - Removing group" << group.id() << "from config file failed";
+            return false;
+        }
+
+        m_groups.erase(it);
+
+        Q_EMIT q->groupRemoved(group);
+
+        return true;
     }
 
 private:
@@ -236,7 +470,9 @@ private:
     bool m_initalized;
     bool m_pgpOnly;
     bool m_remarks_enabled;
-    QMap <QString, std::vector<Key> > m_groups;
+    bool m_groupsEnabled = false;
+    QString m_groupsConfigName;
+    std::vector<KeyGroup> m_groups;
 };
 
 std::shared_ptr<const KeyCache> KeyCache::instance()
@@ -264,6 +500,19 @@ KeyCache::KeyCache()
 
 KeyCache::~KeyCache() {}
 
+void KeyCache::setGroupsEnabled(bool enabled)
+{
+    d->m_groupsEnabled = enabled;
+    if (d->m_initalized) {
+        d->updateGroupCache();
+    }
+}
+
+void KeyCache::setGroupsConfig(const QString &filename)
+{
+    d->m_groupsConfigName = filename;
+}
+
 void KeyCache::enableFileSystemWatcher(bool enable)
 {
     for (const auto &i : qAsConst(d->m_fsWatchers)) {
@@ -463,20 +712,29 @@ std::vector<Key> KeyCache::findByKeyIDOrFingerprint(const std::vector<std::strin
     return result;
 }
 
-const Subkey &KeyCache::findSubkeyByKeyGrip(const char *grip) const
+const Subkey &KeyCache::findSubkeyByKeyGrip(const char *grip, Protocol protocol) const
 {
-    const std::vector<Subkey>::const_iterator it = d->find_keygrip(grip);
-    if (it == d->by.keygrip.end()) {
-        static const Subkey null;
+    static const Subkey null;
+    d->ensureCachePopulated();
+    const auto range = std::equal_range(d->by.keygrip.begin(), d->by.keygrip.end(),
+                                        grip, _detail::ByKeyGrip<std::less>());
+    if (range.first == d->by.keygrip.end()) {
         return null;
+    } else if (protocol == UnknownProtocol) {
+        return *range.first;
     } else {
-        return *it;
+        for (auto it = range.first; it != range.second; ++it) {
+            if (it->parent().protocol() == protocol) {
+                return *it;
+            }
+        }
     }
+    return null;
 }
 
-const Subkey &KeyCache::findSubkeyByKeyGrip(const std::string &grip) const
+const Subkey &KeyCache::findSubkeyByKeyGrip(const std::string &grip, Protocol protocol) const
 {
-    return findSubkeyByKeyGrip(grip.c_str());
+    return findSubkeyByKeyGrip(grip.c_str(), protocol);
 }
 
 std::vector<Subkey> KeyCache::findSubkeysByKeyID(const std::vector<std::string> &ids) const
@@ -502,7 +760,8 @@ std::vector<Subkey> KeyCache::findSubkeysByKeyID(const std::vector<std::string>
 std::vector<Key> KeyCache::findRecipients(const DecryptionResult &res) const
 {
     std::vector<std::string> keyids;
-    Q_FOREACH (const DecryptionResult::Recipient &r, res.recipients())
+    const auto recipients = res.recipients();
+    for (const DecryptionResult::Recipient &r : recipients)
         if (const char *kid = r.keyID()) {
             keyids.push_back(kid);
         }
@@ -520,7 +779,8 @@ std::vector<Key> KeyCache::findRecipients(const DecryptionResult &res) const
 std::vector<Key> KeyCache::findSigners(const VerificationResult &res) const
 {
     std::vector<std::string> fprs;
-    Q_FOREACH (const Signature &s, res.signatures())
+    const auto signatures = res.signatures();
+    for (const Signature &s : signatures)
         if (const char *fpr = s.fingerprint()) {
             fprs.push_back(fpr);
         }
@@ -758,7 +1018,8 @@ static std::string email(const UserID &uid)
 static std::vector<std::string> emails(const Key &key)
 {
     std::vector<std::string> emails;
-    Q_FOREACH (const UserID &uid, key.userIDs()) {
+    const auto userIDs = key.userIDs();
+    for (const UserID &uid : userIDs) {
         const std::string e = email(uid);
         if (!e.empty()) {
             emails.push_back(e);
@@ -868,6 +1129,111 @@ std::vector<Key> KeyCache::secretKeys() const
     return keys;
 }
 
+std::vector<KeyGroup> KeyCache::groups() const
+{
+    d->ensureCachePopulated();
+    return d->m_groups;
+}
+
+std::vector<KeyGroup> KeyCache::configurableGroups() const
+{
+    std::vector<KeyGroup> groups;
+    groups.reserve(d->m_groups.size());
+    std::copy_if(d->m_groups.cbegin(), d->m_groups.cend(),
+                 std::back_inserter(groups),
+                 [] (const KeyGroup &group) { return group.source() == KeyGroup::ApplicationConfig; });
+    return groups;
+}
+
+namespace
+{
+bool compareById(const KeyGroup &lhs, const KeyGroup &rhs)
+{
+    return lhs.id() < rhs.id();
+}
+
+std::vector<KeyGroup> sortedById(std::vector<KeyGroup> groups)
+{
+    std::sort(groups.begin(), groups.end(), &compareById);
+    return groups;
+}
+}
+
+void KeyCache::saveConfigurableGroups(const std::vector<KeyGroup> &groups)
+{
+    const std::vector<KeyGroup> oldGroups = sortedById(configurableGroups());
+    const std::vector<KeyGroup> newGroups = sortedById(groups);
+
+    {
+        std::vector<KeyGroup> removedGroups;
+        std::set_difference(oldGroups.begin(), oldGroups.end(),
+                            newGroups.begin(), newGroups.end(),
+                            std::back_inserter(removedGroups),
+                            &compareById);
+        for (const auto &group : qAsConst(removedGroups)) {
+            qCDebug(LIBKLEO_LOG) << "Removing group" << group;
+            d->remove(group);
+        }
+    }
+    {
+        std::vector<KeyGroup> updatedGroups;
+        std::set_intersection(newGroups.begin(), newGroups.end(),
+                              oldGroups.begin(), oldGroups.end(),
+                              std::back_inserter(updatedGroups),
+                              &compareById);
+        for (const auto &group : qAsConst(updatedGroups)) {
+            qCDebug(LIBKLEO_LOG) << "Updating group" << group;
+            d->update(group);
+        }
+    }
+    {
+        std::vector<KeyGroup> addedGroups;
+        std::set_difference(newGroups.begin(), newGroups.end(),
+                            oldGroups.begin(), oldGroups.end(),
+                            std::back_inserter(addedGroups),
+                            &compareById);
+        for (const auto &group : qAsConst(addedGroups)) {
+            qCDebug(LIBKLEO_LOG) << "Adding group" << group;
+            d->insert(group);
+        }
+    }
+
+    Q_EMIT keysMayHaveChanged();
+}
+
+bool KeyCache::insert(const KeyGroup &group)
+{
+    if (!d->insert(group)) {
+        return false;
+    }
+
+    Q_EMIT keysMayHaveChanged();
+
+    return true;
+}
+
+bool KeyCache::update(const KeyGroup &group)
+{
+    if (!d->update(group)) {
+        return false;
+    }
+
+    Q_EMIT keysMayHaveChanged();
+
+    return true;
+}
+
+bool KeyCache::remove(const KeyGroup &group)
+{
+    if (!d->remove(group)) {
+        return false;
+    }
+
+    Q_EMIT keysMayHaveChanged();
+
+    return true;
+}
+
 void KeyCache::refresh(const std::vector<Key> &keys)
 {
     // make this better...
@@ -1266,6 +1632,36 @@ static bool subkeyIsOk(const Subkey &s)
     return !s.isRevoked() && !s.isInvalid() && !s.isDisabled();
 }
 
+namespace
+{
+time_t creationTimeOfNewestSuitableSubKey(const Key &key, bool needSign, bool needEncrypt)
+{
+    time_t creationTime = 0;
+    for (const Subkey &s: key.subkeys()) {
+        if (!subkeyIsOk(s)) {
+            continue;
+        }
+        if (needSign && !s.canSign()) {
+            continue;
+        }
+        if (needEncrypt && !s.canEncrypt()) {
+            continue;
+        }
+        if (s.creationTime() > creationTime) {
+            creationTime = s.creationTime();
+        }
+    }
+    return creationTime;
+}
+
+struct BestMatch
+{
+    Key key;
+    UserID uid;
+    time_t creationTime = 0;
+};
+}
+
 std::vector<GpgME::Key> KeyCache::findBestByMailBox(const char *addr, GpgME::Protocol proto,
                                                     bool needSign, bool needEncrypt) const
 {
@@ -1282,9 +1678,14 @@ std::vector<GpgME::Key> KeyCache::findBestByMailBox(const char *addr, GpgME::Pro
             return ret;
         }
     }
-    Key keyC;
-    UserID uidC;
-    for (const Key &k: findByEMailAddress(addr)) {
+    // support lookup of email addresses enclosed in angle brackets
+    QByteArray address(addr);
+    if (address[0] == '<' && address[address.size() - 1] == '>') {
+        address = address.mid(1, address.size() - 2);
+    }
+    address = address.toLower();
+    BestMatch best;
+    for (const Key &k: findByEMailAddress(address.constData())) {
         if (proto != Protocol::UnknownProtocol && k.protocol() != proto) {
             continue;
         }
@@ -1294,60 +1695,50 @@ std::vector<GpgME::Key> KeyCache::findBestByMailBox(const char *addr, GpgME::Pro
         if (needSign && (!k.canSign() || !k.hasSecret())) {
             continue;
         }
-        /* First get the uid that matches the mailbox */
+        const time_t creationTime = creationTimeOfNewestSuitableSubKey(k, needSign, needEncrypt);
+        if (creationTime == 0) {
+            // key does not have a suitable (and usable) subkey
+            continue;
+        }
         for (const UserID &u: k.userIDs()) {
-            if (QString::fromStdString(u.addrSpec()).toLower() != QString::fromUtf8(addr).toLower()) {
+            if (QByteArray::fromStdString(u.addrSpec()).toLower() != address) {
+                // user ID does not match the given email address
                 continue;
             }
-            if (uidC.isNull()) {
-                keyC = k;
-                uidC = u;
-            } else if ((!uidIsOk(uidC) && uidIsOk(u)) || uidC.validity() < u.validity()) {
-                /* Validity of the new key is better. */
-                uidC = u;
-                keyC = k;
-            } else if (uidC.validity() == u.validity() && uidIsOk(u)) {
-                /* Both are the same check which one is newer. */
-                time_t oldTime = 0;
-                for (const Subkey &s: keyC.subkeys()) {
-                    if (!subkeyIsOk(s)) {
-                        continue;
-                    }
-                    if (needSign && s.canSign()) {
-                        oldTime = s.creationTime();
-                    } else if (needEncrypt && s.canEncrypt()) {
-                        oldTime = s.creationTime();
-                    } else if (s.canSign() || s.canEncrypt()) {
-                        oldTime = s.creationTime();
-                    }
-                }
-                time_t newTime = 0;
-                for (const Subkey &s: k.subkeys()) {
-                    if (!subkeyIsOk(s)) {
-                        continue;
-                    }
-                    if (needSign && s.canSign()) {
-                        newTime = s.creationTime();
-                    } else if (needEncrypt && s.canEncrypt()) {
-                        newTime = s.creationTime();
-                    } else if (s.canSign() || s.canEncrypt()) {
-                        newTime = s.creationTime();
-                    }
-                }
-                if (newTime > oldTime) {
-                    uidC = u;
-                    keyC = k;
-                }
+            if (best.uid.isNull()) {
+                // we have found our first candidate
+                best = {k, u, creationTime};
+            } else if (!uidIsOk(best.uid) && uidIsOk(u)) {
+                // validity of the new key is better
+                best = {k, u, creationTime};
+            } else if (!k.isExpired() && best.uid.validity() < u.validity()) {
+                // validity of the new key is better
+                best = {k, u, creationTime};
+            } else if (best.key.isExpired() && !k.isExpired()) {
+                // validity of the new key is better
+                best = {k, u, creationTime};
+            } else if (best.uid.validity() == u.validity() && uidIsOk(u) && best.creationTime < creationTime) {
+                // both keys/user IDs have same validity, but the new key is newer
+                best = {k, u, creationTime};
             }
         }
     }
-    ret.push_back(keyC);
+    ret.push_back(best.key);
     return ret;
 }
 
-std::vector<GpgME::Key> KeyCache::getGroupKeys(const QString &groupName) const
+std::vector<Key> KeyCache::getGroupKeys(const QString &groupName) const
 {
-    return d->m_groups.value(groupName);
+    std::vector<Key> result;
+    for (const KeyGroup &g : qAsConst(d->m_groups)) {
+        if (g.name() == groupName) {
+            const KeyGroup::Keys &keys = g.keys();
+            std::copy(keys.cbegin(), keys.cend(), std::back_inserter(result));
+        }
+    }
+    _detail::sort_by_fpr(result);
+    _detail::remove_duplicates_by_fpr(result);
+    return result;
 }
 
 #include "moc_keycache_p.cpp"
