@@ -1,14 +1,19 @@
 /* -*- mode: c++; c-basic-offset:4 -*-
     models/keylistmodel.cpp
 
-    This file is part of Kleopatra, the KDE keymanager
+    This file is part of libkleopatra, the KDE keymanagement library
     SPDX-FileCopyrightText: 2007 Klarälvdalens Datakonsult AB
+    SPDX-FileCopyrightText: 2021 g10 Code GmbH
+    SPDX-FileContributor: Ingo Klöcker <dev@ingo-kloecker.de>
 
     SPDX-License-Identifier: GPL-2.0-or-later
 */
 
 #include "keylistmodel.h"
+
 #include "keycache.h"
+#include "keylist.h"
+#include "kleo/keygroup.h"
 #include "kleo/predicates.h"
 #include "kleo/keyfiltermanager.h"
 #include "kleo/keyfilter.h"
@@ -47,21 +52,45 @@
 
 using namespace GpgME;
 using namespace Kleo;
+using namespace Kleo::KeyList;
 
 Q_DECLARE_METATYPE(GpgME::Key)
+Q_DECLARE_METATYPE(KeyGroup)
 
 class AbstractKeyListModel::Private
 {
+    AbstractKeyListModel *const q;
+public:
+    explicit Private(AbstractKeyListModel *qq);
+
+    void updateFromKeyCache();
+
 public:
     int m_toolTipOptions = Formatting::Validity;
     mutable QHash<const char *, QVariant> prettyEMailCache;
     mutable QHash<const char *, QVariant> remarksCache;
     bool m_useKeyCache = false;
-    bool m_secretOnly = false;
+    KeyList::Options m_keyListOptions = AllKeys;
     std::vector<GpgME::Key> m_remarkKeys;
 };
+
+AbstractKeyListModel::Private::Private(Kleo::AbstractKeyListModel *qq)
+    : q(qq)
+{
+}
+
+void AbstractKeyListModel::Private::updateFromKeyCache()
+{
+    if (m_useKeyCache) {
+        q->setKeys(m_keyListOptions == SecretKeysOnly ? KeyCache::instance()->secretKeys() : KeyCache::instance()->keys());
+        if (m_keyListOptions == IncludeGroups) {
+            q->setGroups(KeyCache::instance()->groups());
+        }
+    }
+}
+
 AbstractKeyListModel::AbstractKeyListModel(QObject *p)
-    : QAbstractItemModel(p), KeyListModelInterface(), d(new Private)
+    : QAbstractItemModel(p), KeyListModelInterface(), d(new Private(this))
 {
 
 }
@@ -106,10 +135,25 @@ std::vector<Key> AbstractKeyListModel::keys(const QList<QModelIndex> &indexes) c
                    [this](const QModelIndex &idx) {
                        return this->key(idx);
                    });
-    result.erase(std::unique(result.begin(), result.end(), _detail::ByFingerprint<std::equal_to>()), result.end());
+    result.erase(std::remove_if(result.begin(), result.end(), std::mem_fn(&GpgME::Key::isNull)), result.end());
+    _detail::remove_duplicates_by_fpr(result);
     return result;
 }
 
+KeyGroup AbstractKeyListModel::group(const QModelIndex &idx) const
+{
+    if (idx.isValid()) {
+        return doMapToGroup(idx);
+    } else {
+        return KeyGroup();
+    }
+}
+
+QModelIndex AbstractKeyListModel::index(const Key &key) const
+{
+    return index(key, 0);
+}
+
 QModelIndex AbstractKeyListModel::index(const Key &key, int col) const
 {
     if (key.isNull() || col < 0 || col >= NumColumns) {
@@ -131,9 +175,23 @@ QList<QModelIndex> AbstractKeyListModel::indexes(const std::vector<Key> &keys) c
     return result;
 }
 
+QModelIndex AbstractKeyListModel::index(const KeyGroup &group) const
+{
+    return index(group, 0);
+}
+
+QModelIndex AbstractKeyListModel::index(const KeyGroup &group, int col) const
+{
+    if (group.isNull() || col < 0 || col >= NumColumns) {
+        return {};
+    } else {
+        return doMapFromGroup(group, col);
+    }
+}
+
 void AbstractKeyListModel::setKeys(const std::vector<Key> &keys)
 {
-    clear();
+    clear(Keys);
     addKeys(keys);
 }
 
@@ -165,12 +223,36 @@ QList<QModelIndex> AbstractKeyListModel::addKeys(const std::vector<Key> &keys)
     return doAddKeys(sorted);
 }
 
-void AbstractKeyListModel::clear()
+void AbstractKeyListModel::setGroups(const std::vector<KeyGroup> &groups)
+{
+    clear(Groups);
+    doSetGroups(groups);
+}
+
+QModelIndex AbstractKeyListModel::addGroup(const KeyGroup &group)
+{
+    if (group.isNull()) {
+        return QModelIndex();
+    }
+    return doAddGroup(group);
+}
+
+bool AbstractKeyListModel::removeGroup(const KeyGroup &group)
+{
+    if (group.isNull()) {
+        return false;
+    }
+    return doRemoveGroup(group);
+}
+
+void AbstractKeyListModel::clear(ItemTypes types)
 {
     beginResetModel();
-    doClear();
-    d->prettyEMailCache.clear();
-    d->remarksCache.clear();
+    doClear(types);
+    if (types & Keys) {
+        d->prettyEMailCache.clear();
+        d->remarksCache.clear();
+    }
     endResetModel();
 }
 
@@ -225,13 +307,21 @@ static QVariant returnIfValid(const QIcon &t)
 QVariant AbstractKeyListModel::data(const QModelIndex &index, int role) const
 {
     const Key key = this->key(index);
-    if (key.isNull()) {
-        return QVariant();
+    if (!key.isNull()) {
+        return data(key, index.column(), role);
     }
 
-    const int column = index.column();
+    const KeyGroup group = this->group(index);
+    if (!group.isNull()) {
+        return data(group, index.column(), role);
+    }
 
-    if (role == Qt::DisplayRole || role == Qt::EditRole) {
+    return QVariant();
+}
+
+QVariant AbstractKeyListModel::data(const Key &key, int column, int role) const
+{
+   if (role == Qt::DisplayRole || role == Qt::EditRole) {
         switch (column) {
         case PrettyName:
             return Formatting::prettyName(key);
@@ -335,6 +425,71 @@ QVariant AbstractKeyListModel::data(const QModelIndex &index, int role) const
     return QVariant();
 }
 
+QVariant AbstractKeyListModel::data(const KeyGroup &group, int column, int role) const
+{
+    if (role == Qt::DisplayRole || role == Qt::EditRole) {
+        switch (column) {
+        case PrettyName:
+            return group.name();
+        case PrettyEMail:
+            return QVariant();
+        case Validity:
+            return Formatting::complianceStringShort(group);
+        case ValidFrom:
+            return QString();
+        case ValidUntil:
+            return QString();
+        case TechnicalDetails:
+            return Formatting::type(group);
+        case ShortKeyID:
+            return QString();
+        case KeyID:
+            return QString();
+        case Summary:
+            return Formatting::summaryLine(group);  // used for filtering
+        case Fingerprint:
+            return QString();
+        case Issuer:
+            return QString();
+        case Origin:
+            return QString();
+        case LastUpdate:
+            return QString();
+        case SerialNumber:
+            return QString();
+        case OwnerTrust:
+            return QString();
+        case Remarks:
+            return QVariant();
+        case NumColumns:
+            break;
+        }
+    } else if (role == Qt::ToolTipRole) {
+        return Formatting::toolTip(group, toolTipOptions());
+    } else if (role == Qt::FontRole) {
+        return QFont();
+    } else if (role == Qt::DecorationRole) {
+        return column == Icon ? QIcon::fromTheme("group") : QVariant();
+    } else if (role == Qt::BackgroundRole) {
+    } else if (role == Qt::ForegroundRole) {
+    } else if (role == GroupRole) {
+        return QVariant::fromValue(group);
+    }
+    return QVariant();
+}
+
+bool AbstractKeyListModel::setData(const QModelIndex &index, const QVariant &value, int role)
+{
+    Q_UNUSED(role)
+    Q_ASSERT(value.canConvert<KeyGroup>());
+    if (value.canConvert<KeyGroup>()) {
+        const KeyGroup group = value.value<KeyGroup>();
+        return doSetGroupData(index, group);
+    }
+
+    return false;
+}
+
 namespace
 {
 template <typename Base>
@@ -375,7 +530,7 @@ public:
 
     int rowCount(const QModelIndex &pidx) const override
     {
-        return pidx.isValid() ? 0 : mKeysByFingerprint.size();
+        return pidx.isValid() ? 0 : mKeysByFingerprint.size() + mGroups.size();
     }
 
 private:
@@ -383,12 +538,47 @@ private:
     QModelIndex doMapFromKey(const Key &key, int col) const override;
     QList<QModelIndex> doAddKeys(const std::vector<Key> &keys) override;
     void doRemoveKey(const Key &key) override;
-    void doClear() override {
-        mKeysByFingerprint.clear();
+
+    KeyGroup doMapToGroup(const QModelIndex &index) const override;
+    QModelIndex doMapFromGroup(const KeyGroup &group, int column) const override;
+    void doSetGroups(const std::vector<KeyGroup> &groups) override;
+    QModelIndex doAddGroup(const KeyGroup &group) override;
+    bool doSetGroupData(const QModelIndex &index, const KeyGroup &group) override;
+    bool doRemoveGroup(const KeyGroup &group) override;
+
+    void doClear(ItemTypes types) override
+    {
+        if (types & Keys) {
+            mKeysByFingerprint.clear();
+        }
+        if (types & Groups) {
+            mGroups.clear();
+        }
+    }
+
+    int firstGroupRow() const
+    {
+        return mKeysByFingerprint.size();
+    }
+
+    int lastGroupRow() const
+    {
+        return mKeysByFingerprint.size() + mGroups.size() - 1;
+    }
+
+    int groupIndex(const QModelIndex &index) const
+    {
+        if (!index.isValid()
+                || index.row() < firstGroupRow() || index.row() > lastGroupRow()
+                || index.column() >= NumColumns) {
+            return -1;
+        }
+        return index.row() - firstGroupRow();
     }
 
 private:
     std::vector<Key> mKeysByFingerprint;
+    std::vector<KeyGroup> mGroups;
 };
 
 class HierarchicalKeyListModel : public AbstractKeyListModel
@@ -413,11 +603,45 @@ private:
     QModelIndex doMapFromKey(const Key &key, int col) const override;
     QList<QModelIndex> doAddKeys(const std::vector<Key> &keys) override;
     void doRemoveKey(const Key &key) override;
-    void doClear() override {
-        mTopLevels.clear();
-        mKeysByFingerprint.clear();
-        mKeysByExistingParent.clear();
-        mKeysByNonExistingParent.clear();
+
+    KeyGroup doMapToGroup(const QModelIndex &index) const override;
+    QModelIndex doMapFromGroup(const KeyGroup &group, int column) const override;
+    void doSetGroups(const std::vector<KeyGroup> &groups) override;
+    QModelIndex doAddGroup(const KeyGroup &group) override;
+    bool doSetGroupData(const QModelIndex &index, const KeyGroup &group) override;
+    bool doRemoveGroup(const KeyGroup &group) override;
+
+    void doClear(ItemTypes types) override
+    {
+        if (types & Keys) {
+            mTopLevels.clear();
+            mKeysByFingerprint.clear();
+            mKeysByExistingParent.clear();
+            mKeysByNonExistingParent.clear();
+        }
+        if (types & Groups) {
+            mGroups.clear();
+        }
+    }
+
+    int firstGroupRow() const
+    {
+        return mTopLevels.size();
+    }
+
+    int lastGroupRow() const
+    {
+        return mTopLevels.size() + mGroups.size() - 1;
+    }
+
+    int groupIndex(const QModelIndex &index) const
+    {
+        if (!index.isValid()
+                || index.row() < firstGroupRow() || index.row() > lastGroupRow()
+                || index.column() >= NumColumns) {
+            return -1;
+        }
+        return index.row() - firstGroupRow();
     }
 
 private:
@@ -430,6 +654,7 @@ private:
     std::vector<Key> mKeysByFingerprint; // all keys
     Map mKeysByExistingParent, mKeysByNonExistingParent; // parent->child map
     std::vector<Key> mTopLevels; // all roots + parent-less
+    std::vector<KeyGroup> mGroups;
 };
 
 static const char *cleanChainID(const Key &key)
@@ -446,10 +671,8 @@ static const char *cleanChainID(const Key &key)
 }
 
 FlatKeyListModel::FlatKeyListModel(QObject *p)
-    : TableModelMixin<AbstractKeyListModel>(p),
-      mKeysByFingerprint()
+    : TableModelMixin<AbstractKeyListModel>(p)
 {
-
 }
 
 FlatKeyListModel::~FlatKeyListModel() {}
@@ -521,6 +744,82 @@ void FlatKeyListModel::doRemoveKey(const Key &key)
     endRemoveRows();
 }
 
+KeyGroup FlatKeyListModel::doMapToGroup(const QModelIndex &idx) const
+{
+    Q_ASSERT(idx.isValid());
+    if (static_cast<unsigned>(idx.row()) >= mKeysByFingerprint.size()
+            && static_cast<unsigned>(idx.row()) < mKeysByFingerprint.size() + mGroups.size()
+            && idx.column() < NumColumns) {
+        return mGroups[ idx.row() - mKeysByFingerprint.size() ];
+    } else {
+        return KeyGroup();
+    }
+}
+
+QModelIndex FlatKeyListModel::doMapFromGroup(const KeyGroup &group, int column) const
+{
+    Q_ASSERT(!group.isNull());
+    const auto it = std::find_if(mGroups.cbegin(), mGroups.cend(),
+                                 [group](const KeyGroup &g) {
+                                     return g.source() == group.source() && g.id() == group.id();
+                                 });
+    if (it == mGroups.cend()) {
+        return QModelIndex();
+    } else {
+        return createIndex(it - mGroups.cbegin() + mKeysByFingerprint.size(), column);
+    }
+}
+
+void FlatKeyListModel::doSetGroups(const std::vector<KeyGroup> &groups)
+{
+    Q_ASSERT(mGroups.empty());  // ensure that groups have been cleared
+    const int first = mKeysByFingerprint.size();
+    const int last = first + groups.size() - 1;
+    beginInsertRows(QModelIndex(), first, last);
+    mGroups = groups;
+    endInsertRows();
+}
+
+QModelIndex FlatKeyListModel::doAddGroup(const KeyGroup &group)
+{
+    const int newRow = lastGroupRow() + 1;
+    beginInsertRows(QModelIndex(), newRow, newRow);
+    mGroups.push_back(group);
+    endInsertRows();
+    return createIndex(newRow, 0);
+}
+
+bool FlatKeyListModel::doSetGroupData(const QModelIndex &index, const KeyGroup &group)
+{
+    if (group.isNull()) {
+        return false;
+    }
+    const int groupIndex = this->groupIndex(index);
+    if (groupIndex == -1) {
+        return false;
+    }
+    mGroups[groupIndex] = group;
+    Q_EMIT dataChanged(createIndex(index.row(), 0), createIndex(index.row(), NumColumns - 1));
+    return true;
+}
+
+bool FlatKeyListModel::doRemoveGroup(const KeyGroup &group)
+{
+    const QModelIndex modelIndex = doMapFromGroup(group, 0);
+    if (!modelIndex.isValid()) {
+        return false;
+    }
+    const int groupIndex = this->groupIndex(modelIndex);
+    Q_ASSERT(groupIndex != -1);
+    if (groupIndex == -1) {
+        return false;
+    }
+    beginRemoveRows(QModelIndex(), modelIndex.row(), modelIndex.row());
+    mGroups.erase(mGroups.begin() + groupIndex);
+    endRemoveRows();
+    return true;
+}
+
 HierarchicalKeyListModel::HierarchicalKeyListModel(QObject *p)
     : AbstractKeyListModel(p),
       mKeysByFingerprint(),
@@ -538,7 +837,7 @@ int HierarchicalKeyListModel::rowCount(const QModelIndex &pidx) const
 
     // toplevel item:
     if (!pidx.isValid()) {
-        return mTopLevels.size();
+        return mTopLevels.size() + mGroups.size();
     }
 
     if (pidx.column() != 0) {
@@ -569,6 +868,8 @@ QModelIndex HierarchicalKeyListModel::index(int row, int col, const QModelIndex
     if (!pidx.isValid()) {
         if (static_cast<unsigned>(row) < mTopLevels.size()) {
             return index(mTopLevels[row], col);
+        } else if (static_cast<unsigned>(row) < mTopLevels.size() + mGroups.size()) {
+            return index(mGroups[row - mTopLevels.size()], col);
         } else {
             return QModelIndex();
         }
@@ -779,7 +1080,8 @@ QList<QModelIndex> HierarchicalKeyListModel::doAddKeys(const std::vector<Key> &k
 
     std::set<Key, _detail::ByFingerprint<std::less> > changedParents;
 
-    Q_FOREACH (const Key &key, topological_sort(keys)) {
+    const auto topologicalSortedList = topological_sort(keys);
+    for (const Key &key : topologicalSortedList) {
 
         // check to see whether this key is a parent for a previously parent-less group:
         const char *const fpr = key.primaryFingerprint();
@@ -882,7 +1184,7 @@ void HierarchicalKeyListModel::doRemoveKey(const Key &key)
         keys.erase(it);
         // FIXME for simplicity, we just clear the model and re-add all keys minus the removed one. This is suboptimal,
         // but acceptable given that deletion of non-leave nodes is rather rare.
-        clear();
+        clear(Keys);
         addKeys(keys);
         return;
     }
@@ -932,16 +1234,98 @@ void HierarchicalKeyListModel::doRemoveKey(const Key &key)
     endRemoveRows();
 }
 
-void AbstractKeyListModel::useKeyCache(bool value, bool secretOnly)
+KeyGroup HierarchicalKeyListModel::doMapToGroup(const QModelIndex &idx) const
 {
-    d->m_secretOnly = secretOnly;
-    d->m_useKeyCache = value;
-    if (value) {
-        setKeys(d->m_secretOnly ? KeyCache::instance()->secretKeys() : KeyCache::instance()->keys());
-    } else {
-        setKeys(std::vector<Key>());
+    Q_ASSERT(idx.isValid());
+    if (idx.parent().isValid()) {
+        // groups are always top-level
+        return KeyGroup();
     }
-    connect(KeyCache::instance().get(), &KeyCache::keysMayHaveChanged, this, [this] { if (d->m_useKeyCache) { setKeys(d->m_secretOnly ? KeyCache::instance()->secretKeys() : KeyCache::instance()->keys()); } });
+
+    if (static_cast<unsigned>(idx.row()) >= mTopLevels.size()
+            && static_cast<unsigned>(idx.row()) < mTopLevels.size() + mGroups.size()
+            && idx.column() < NumColumns) {
+        return mGroups[ idx.row() - mTopLevels.size() ];
+    } else {
+        return KeyGroup();
+    }
+}
+
+QModelIndex HierarchicalKeyListModel::doMapFromGroup(const KeyGroup &group, int column) const
+{
+    Q_ASSERT(!group.isNull());
+    const auto it = std::find_if(mGroups.cbegin(), mGroups.cend(),
+                                 [group](const KeyGroup &g) {
+                                     return g.source() == group.source() && g.id() == group.id();
+                                 });
+    if (it == mGroups.cend()) {
+        return QModelIndex();
+    } else {
+        return createIndex(it - mGroups.cbegin() + mTopLevels.size(), column);
+    }
+}
+
+void HierarchicalKeyListModel::doSetGroups(const std::vector<KeyGroup> &groups)
+{
+    Q_ASSERT(mGroups.empty());  // ensure that groups have been cleared
+    const int first = mTopLevels.size();
+    const int last = first + groups.size() - 1;
+    beginInsertRows(QModelIndex(), first, last);
+    mGroups = groups;
+    endInsertRows();
+}
+
+QModelIndex HierarchicalKeyListModel::doAddGroup(const KeyGroup &group)
+{
+    const int newRow = lastGroupRow() + 1;
+    beginInsertRows(QModelIndex(), newRow, newRow);
+    mGroups.push_back(group);
+    endInsertRows();
+    return createIndex(newRow, 0);
+}
+
+bool HierarchicalKeyListModel::doSetGroupData(const QModelIndex &index, const KeyGroup &group)
+{
+    if (group.isNull()) {
+        return false;
+    }
+    const int groupIndex = this->groupIndex(index);
+    if (groupIndex == -1) {
+        return false;
+    }
+    mGroups[groupIndex] = group;
+    Q_EMIT dataChanged(createIndex(index.row(), 0), createIndex(index.row(), NumColumns - 1));
+    return true;
+}
+
+bool HierarchicalKeyListModel::doRemoveGroup(const KeyGroup &group)
+{
+    const QModelIndex modelIndex = doMapFromGroup(group, 0);
+    if (!modelIndex.isValid()) {
+        return false;
+    }
+    const int groupIndex = this->groupIndex(modelIndex);
+    Q_ASSERT(groupIndex != -1);
+    if (groupIndex == -1) {
+        return false;
+    }
+    beginRemoveRows(QModelIndex(), modelIndex.row(), modelIndex.row());
+    mGroups.erase(mGroups.begin() + groupIndex);
+    endRemoveRows();
+    return true;
+}
+
+void AbstractKeyListModel::useKeyCache(bool value, KeyList::Options options)
+{
+    d->m_keyListOptions = options;
+    d->m_useKeyCache = value;
+    if (!d->m_useKeyCache) {
+        clear(All);
+    } else {
+        d->updateFromKeyCache();
+    }
+    connect(KeyCache::instance().get(), &KeyCache::keysMayHaveChanged,
+            this, [this] { d->updateFromKeyCache(); });
 }
 
 // static
