@@ -296,6 +296,8 @@ public:
 
     }
 
+    ~Private() = default;
+
     void generateKey(KeySelectionCombo *combo)
     {
         const auto &addr = combo->property("address").toString();
@@ -689,8 +691,6 @@ public:
         }
     }
 
-    ~Private() {}
-
     GpgME::Protocol mProto;
     QList<KeySelectionCombo *> mSigningCombos;
     QList<KeySelectionCombo *> mEncCombos;
@@ -735,6 +735,8 @@ NewKeyApprovalDialog::NewKeyApprovalDialog(const QMap<QString, std::vector<GpgME
     resize(QSize(desk.width() / 3, qMin(size.height(), desk.height() / 2)));
 }
 
+Kleo::NewKeyApprovalDialog::~NewKeyApprovalDialog() = default;
+
 std::vector<GpgME::Key> NewKeyApprovalDialog::signingKeys()
 {
     return d->mAcceptedSig;
